
#include <stdio.h>
#include <stdlib.h>
#include "cell.h"
#include "rif.h"

#define LIMIT	1e-9
#define WIDTH	596.0
#define HEIGHT	842.0
#define FOLD	36

static void	Error(char *msg)
{
	(void)fputs(msg,stderr); (void)fputc('\n',stderr); exit(1);
}

typedef struct List {
		int		x,y;
		struct List	*next;
	} List;

static List	*Alloc(int x,int y,List *next)
{
	List	*list=(List *)malloc(sizeof(*list));

	if (list==NULL) Error("list memory allocation error.");

	list->x=x; list->y=y; list->next=next; return list;
}

static List	*Trace(List *list)
{
	List	*next;

	while ((next=list->next)!=NULL) list=next; return list;
}

static char	form[4+4+1],*lw;
static int	fold;

static void	Print(int No,List *list)
{
	int	no;

	if (lw!=NULL) {
	    (void)printf("%s setlinewidth\n",lw);

	    lw=NULL;
	}
	(void)printf("%d D\n",No);
	for (no=0; ++no<=No; list=list->next) {
	    (void)printf(form,list->x,list->y);
	    if (no==No || no%fold==0) (void)putchar('\n');
	}
}

static int	Nx,Ny;

static void	Flush(List *list)
{
	List	*tmp,*top=list;
	int	no;

#define OB	(list->x==0 || list->x==Nx || list->y==0 || list->y==Ny)

	if (!OB) {
	    (tmp=Trace(list))->next=list->next;

	    while ((list=list->next)!=tmp)
		if (OB) {
		    *top=(*list); break;
		}

	    list->next=NULL; list=top;
	}
	for (no=0; list!=NULL; list=list->next)
	    if (OB) {
		if (no>1) Print(no+1,tmp);

		no=1; tmp=list;
	    }
	    else
		if (list->next==NULL)
		    Print(no+1,tmp);
		else
		    if (++no==1) tmp=list;

	while ((list=top)!=NULL) {
	    top=list->next; free(list);
	}
}

static int	Na,Nb,M;
static Cell	**grid,g1,g2;
static char	*last;
static List	**vert;

static void	ContourLine(double g)
{
	int	x,y,b2,b1,code,a2,a1;
	double	b,a,g0,ga,gb;
	char	flag;
	List	*hori,*temp;

	if (g<(double)g1 || g>(double)g2) return;

	for (x=0; x<=Nx; x++) last[x]=0;

	for (y=0; y<=Ny; y++) {
	    if (y!=Ny)
		if ((b2=(int)((b=((double)y+0.5)/(double)M)+0.5))==0) {
		    b1=0; b+=0.5;
		}
		else {
		    if (b2==Nb) b2=b1=Nb-1; else b1=b2-1;

		    b-=((double)b1+0.5);
		}

	    code=0;
	    for (x=0; x<=Nx; x++) {
		if (x==Nx || y==Ny)
		    flag=0;
		else {
		    if ((a2=(int)((a=((double)x+0.5)/(double)M)+0.5))==0) {
			a1=0; a+=0.5;
		    }
		    else {
			if (a2==Na) a2=a1=Na-1; else a1=a2-1;

			a-=((double)a1+0.5);
		    }
		    g0=(double)grid[b1][a1];
		    ga=(double)grid[b1][a2]-g0;
		    gb=(double)grid[b2][a1]-g0;
		    flag=g>=g0+ga*a+gb*b+((double)grid[b2][a2]-g0-ga-gb)*a*b;
		}
		code=((code&3)<<2)|(last[x]<<1)|flag; last[x]=flag;

#define VERT	vert[x]
#define X	x+1
#define Y	y+1

		switch(code) {
		    case 0x1:
			VERT=Alloc(x,Y,Alloc(x,y,hori=Alloc(X,y,NULL)));
			break;
		    case 0x2:
			VERT->y=y; hori=Alloc(X,y,VERT);
			break;
		    case 0x4:
			hori->x=x; hori->next=VERT=Alloc(x,Y,NULL);
			break;
		    case 0x6:
			VERT->y=y; temp=Alloc(X,y,VERT);
			hori->x=x; hori->next=VERT=Alloc(x,Y,NULL);
			hori=temp;
			break;
		    case 0x7:
			hori->x=x;
			if (Trace(VERT)==hori) {
			    VERT->y=y; Flush(VERT);
			}
			else {
			    hori->next=VERT->next; free(VERT);
			}
			break;
		    case 0x8:
			VERT->y=y;
			if (Trace(hori)==VERT) {
			    hori->x=x; Flush(hori);
			}
			else {
			    VERT->next=hori->next; free(hori);
			}
			break;
		    case 0x9:
			VERT->y=y;
			if (Trace(hori)==VERT) {
			    hori->x=x; Flush(hori);
			}
			else {
			    VERT->next=hori->next; free(hori);
			}
			VERT=Alloc(x,Y,Alloc(x,y,hori=Alloc(X,y,NULL)));
			break;
		    case 0xb:
			hori->x=x; VERT=Alloc(x,Y,hori);
			break;
		    case 0xd:
			VERT->y=y; VERT->next=hori=Alloc(X,y,NULL);
			break;
		    case 0xe:
			hori=Alloc(X,y,Alloc(x,y,VERT=Alloc(x,Y,NULL)));
			break;
		}
	    }
	}
}

#define LEN	1024

int	main(int argc,char **argv)
{
	double	Sx,Sy,w,h,l,d,r,u,base,step,W=WIDTH,H=HEIGHT;
	int	x,y,b,a,Nc,c;
	char	text[LEN],str[3][LEN];

	if (argc!=4 && argc!=5 && argc!=6 && argc!=7)
	    Error("usage : t2cl TIFF DPI|(xDPI yDPI) multiple {width height}");

	ReadImageFile(argv[1],&Na,&Nb,NULL,&grid,NULL);

	if (argc==4 || argc==6) {
	    Sx=Sy=atof(argv[2]); M=atoi(argv[3]);
	}
	else {
	    Sx=atof(argv[2]);
	    Sy=atof(argv[3]); M=atoi(argv[4]);
	}
	if ((Sx>=(-LIMIT) && Sx<=LIMIT) ||
	    (Sy>=(-LIMIT) && Sy<=LIMIT)) Error("too small imaging DPI.");

	if (M<1) Error("bad multiple for BLI.");

	Sx=72.0/Sx/(double)M; Nx=Na*M;
	Sy=72.0/Sy/(double)M; Ny=Nb*M;

	if ((last=(char  *)malloc((Nx+1)*sizeof(*last)))==NULL ||
	    (vert=(List **)malloc((Nx+1)*sizeof(*vert)))==NULL)
	    Error("scanline memory allocation error.");

#ifdef	LANDSCAPE
	w=((Sy>0.0)?Sy:-Sy)*(double)Ny;
	h=((Sx>0.0)?Sx:-Sx)*(double)Nx;
#else
	w=((Sx>0.0)?Sx:-Sx)*(double)Nx;
	h=((Sy>0.0)?Sy:-Sy)*(double)Ny;
#endif
	if (argc==6 ||argc==7) {
	    W=(argv[argc-2][0]=='-')?(argv[argc-2][1]=='\0')?
	      w:atof(argv[argc-2]+1):atof(argv[argc-2])*72.0;
	    H=(argv[argc-1][0]=='-')?(argv[argc-1][1]=='\0')?
	      h:atof(argv[argc-1]+1):atof(argv[argc-1])*72.0;
	}
	l=(W-w)/2.0; d=(H-h)/2.0; r=(W+w)/2.0; u=(H+h)/2.0;

	for (x=1; Nx>>x*8; x++) ;
	for (y=1; Ny>>y*8; y++) ;

	(void)sprintf(form,"%%0%dx%%0%dx",x*2,y*2); fold=FOLD/(x+y);

	(void)printf("%%!\n"
		     "%%%%BoundingBox: %lf %lf %lf %lf\n"
		     "%%\n"
		     "4 dict begin gsave\n"
		     "/X %d string def\n"
		     "/Y %d string def\n"
		     "/C { 0 currentfile X readhexstring pop"
		         " { exch 256 mul add } forall\n"
		     "     %le mul %lf add\n"
		     "     0 currentfile Y readhexstring pop"
			 " { exch 256 mul add } forall\n"
		     "     %le mul %lf add"
#ifdef	LANDSCAPE    
			 " exch"
#endif
			 " } bind def\n"
		     "/D { newpath C moveto"
		         " 1 sub { C lineto } repeat stroke } bind def\n",
		     l,d,r,u,x,y,
#ifdef	LANDSCAPE
		     Sx,(Sx>0.0)?d:u, Sy,(Sy>0.0)?l:r
#else
		     Sx,(Sx>0.0)?l:r,-Sy,(Sy<0.0)?d:u
#endif
		    );

	g1=g2=grid[0][0];
	for (b=0; b<Nb; b++)
	for (a=0; a<Na; a++)
	    if (grid[b][a]<g1) g1=grid[b][a]; else
	    if (grid[b][a]>g2) g2=grid[b][a];

	while (fgets(text,LEN,stdin)!=NULL) {
	    switch(sscanf(text,"%lf %s %s %s",&base,str[0],str[1],str[2])) {
		case 1 : step=0.0; Nc=1; lw=NULL; break;
		case 2 : step=0.0; Nc=1; lw=str[0]; break;
		case 3 : step=atof(str[0]); Nc=atoi(str[1]); lw=NULL; break;
		case 4 : step=atof(str[0]); Nc=atoi(str[1]); lw=str[2]; break;
		default: continue;
	    }
	    for (c=0; c<Nc; c++) ContourLine(base+step*(double)c);
	}
	(void)printf("grestore end\n"
		     "%%\n"
		     "%%%%%%[ %lf %lf %lf %lf ] rectstroke\n",l,d,w,h);

	free(vert); free(last);

	for (b=Nb-1; b>=0; b--) free(grid[b]); free(grid);

	return 0;
}
