
#include <stdio.h>
#include <stdlib.h>
#include "cell.h"
#include "rif_rgb.h"

#define LIMIT	1e-9	/* DPI lmit */
#define WIDTH	596.0
#define HEIGHT	842.0
#define FOLD	36	/* must be multiplier of 3 */

static void	Error(char *msg)
{
	(void)fputs(msg,stderr); (void)fputc('\n',stderr); exit(1);
}

int	main(int argc,char **argv)
{
	int	Nx,Ny,BPS,y,x,gray=1,fold=0;
	Cell	**R,**G,**B;
	double	Ix,Iy,w,h,l,d,ratio,W=WIDTH,H=HEIGHT;

	if (argc!=3 && argc!=4 && argc!=5 && argc!=6)
	    Error("usage : t2ps TIFF DPI|(xDPI yDPI) {width height}");

	ReadImageFile_RGB(argv[1],&Nx,&Ny,&BPS,&R,&G,&B,NULL);

	if (argc==3 || argc==5)
	    Ix=Iy=atof(argv[2]);
	else {
	    Ix=atof(argv[2]);
	    Iy=atof(argv[3]);
	}
	if ((Ix>=(-LIMIT) && Ix<=LIMIT) ||
	    (Iy>=(-LIMIT) && Iy<=LIMIT)) Error("too small imaging DPI.");

	Ix=((double)Nx*72.0/Ix);
	Iy=((double)Ny*72.0/Iy);

#ifdef	LANDSCAPE
	w=(Iy>0.0)?Iy:-Iy;
	h=(Ix>0.0)?Ix:-Ix;
#else
	w=(Ix>0.0)?Ix:-Ix;
	h=(Iy>0.0)?Iy:-Iy;
#endif
	if (argc==5 || argc==6) {
	    W=(argv[argc-2][0]=='-')?(argv[argc-2][1]=='\0')?
	      w:atof(argv[argc-2]+1):atof(argv[argc-2])*72.0;
	    H=(argv[argc-1][0]=='-')?(argv[argc-1][1]=='\0')?
	      h:atof(argv[argc-1]+1):atof(argv[argc-1])*72.0;
	}
	ratio=255.0/(double)((1<<BPS)-1);
	for (y=0; y<Ny; y++)
	for (x=0; x<Nx; x++) {
	    R[y][x]=(Cell)((double)R[y][x]*ratio+0.5);
	    G[y][x]=(Cell)((double)G[y][x]*ratio+0.5);
	    B[y][x]=(Cell)((double)B[y][x]*ratio+0.5);
	    if (gray) gray=R[y][x]==G[y][x] && G[y][x]==B[y][x];
	}
	(void)printf("%%!\n"
		     "%%%%BoundingBox: %lf %lf %lf %lf\n"
		     "%%\n"
		     "1 dict begin gsave\n"
		     "/HS %d string def\n"
#ifdef	LANDSCAPE
		     "[ 0 %lf %lf 0 %lf %lf ] concat\n"
#else
		     "[ %lf 0 0 %lf %lf %lf ] concat\n"
#endif
		     "%d %d 8 [ %d 0 0 %d 0 0 ]\n"
		     "{ currentfile HS readhexstring pop } bind\n",
		     l=(W-w)/2.0,d=(H-h)/2.0,(W+w)/2.0,(H+h)/2.0,
		     (gray)?Nx:Nx*3,
#ifdef	LANDSCAPE
		     Ix, Iy,(W-Iy)/2.0,(H-Ix)/2.0,
#else
		     Ix,-Iy,(W-Ix)/2.0,(H+Iy)/2.0,
#endif
		     Nx,Ny,Nx,Ny);

	if (gray) {
	    (void)puts("image");
	    for (y=0; y<Ny; y++)
	    for (x=0; x<Nx; x++) {
		(void)printf("%02x",R[y][x]);
		if ((fold=(fold+1)%FOLD)==0) (void)putchar('\n');
	    }
	}
	else {
	    (void)puts("false 3 colorimage");
	    for (y=0; y<Ny; y++)
	    for (x=0; x<Nx; x++) {
		(void)printf("%02x%02x%02x",R[y][x],G[y][x],B[y][x]);
		if ((fold=(fold+3)%FOLD)==0) (void)putchar('\n');
	    }
	}
	if (fold!=0) (void)putchar('\n');

	(void)printf("grestore end\n"
		     "%%\n"
		     "%%%%%%[ %lf %lf %lf %lf ] rectstroke\n",l,d,w,h);

	return 0;
}
